import os


class Paths:
    def __init__(self, project_dir):
        self.project_dir = project_dir

    @property
    def imported_data(self):
        return os.path.join(self.project_dir, "bblocks", "import_tools", "stored_data")


PATHS = Paths(os.path.dirname(os.path.dirname(__file__)))



from dataclasses import dataclass, field

from country_converter import convert


@dataclass(repr=False)
class Dict(dict):
    dictionary: dict = field(default_factory=dict)

    """A wrapper that adds functionality to a standard dictionary"""

    def __post_init__(self):
        self.update(self.dictionary)

    def change_keys(self, from_: str = None, to: str = "ISO3") -> dict:
        _ = {
            convert(names=key, src=from_, to=to, not_found=None): value
            for key, value in self.dictionary.items()
        }
        self.clear()
        self.update(_)
        return self

    def reverse(self) -> dict:
        _ = {value: key for key, value in self.items()}
        self.clear()
        self.update(_)
        return self

    def set_keys_type(self, type_: type) -> dict:
        _ = {type_(key): value for key, value in self.dictionary.items()}
        self.clear()
        self.update(_)
        return self

    def set_values_type(self, type_: type) -> dict:
        _ = {key: type_(value) for key, value in self.dictionary.items()}
        self.clear()
        self.update(_)
        return self



from __future__ import annotations

import pandas as pd
import requests
from datetime import datetime

import weo.dates
from bs4 import BeautifulSoup
import os
from typing import Optional
import warnings

from weo import all_releases, download, WEO

from bblocks.config import PATHS
from bblocks.cleaning_tools.clean import clean_numeric_series
from bblocks.import_tools.common import ImportData


def _parse_sdr_links(url: str, concat_url: str) -> dict:
    """Function to parse SDR tables from html

    When called it finds the SDR tables and finds a relevant
    date and link

    Args:
        url (str): URL to parse
        concat_url: base url onto which a href is attached

    Returns:
        A dictionary with dates as keys and URLs as values
    """

    try:
        content = requests.get(url).content
    except ConnectionError:
        raise ConnectionError(f"Could not read page: {url}")

    soup = BeautifulSoup(content, "html.parser")
    table = soup.find_all("table")[4]
    links = table.find_all("a")

    return {link.string: f'{concat_url}{link.get("href")}' for link in links}


def _get_tsv_url(url: str) -> str:
    """retrieves the URL for the SDR TSV file

    Args:
        url: URL to parse

    Returns:
        A link to the TSV file
    """

    try:
        content = requests.get(url).content
    except ConnectionError:
        raise ConnectionError(f"could not read page: {url}")
    soup = BeautifulSoup(content, "html.parser")
    href = soup.find_all("a", string="TSV")[0].get("href")

    return f"https://www.imf.org/external/np/fin/tad/{href}"


def _read_sdr_tsv(url: str) -> pd.DataFrame:
    """Reads a TSV file with SDR data from the web to a pandas dataframe

    Args:
        url: link to the TSV file

    Returns:
        A pandas dataframe with SDR data
    """

    df = pd.read_csv(url, delimiter="/t", engine="python").loc[3:]
    df = df["SDR Allocations and Holdings"].str.split("\t", expand=True)
    df.columns = ["member", "holdings", "allocations"]

    return (
        df.melt(id_vars="member", value_vars=["holdings", "allocations"])
        .pipe(clean_numeric_series, series_columns="value")
        .rename(columns={"variable": "indicator"})
        .reset_index(drop=True)
    )


def _check_sdr_indicators(indicator: str | list) -> list:
    """Checks if the indicator is in the list of SDR accepted indicators

    Args:
        indicator: indicator to check. If indicator is 'all', all accepted indicators
            are returned ['holdings', 'allocations']

    Returns:
        A list of accepted indicators
    """

    accepted_indicators = ["allocations", "holdings"]

    if indicator == "all":
        return accepted_indicators

    else:
        if isinstance(indicator, str):
            indicator = [indicator]

        for i in indicator:
            if i not in accepted_indicators:
                raise ValueError(f"{i} is not a valid indicator.")

        return indicator


def _get_data(obj: ImportData, indicators: str | list) -> pd.DataFrame:
    """Unpack dictionary of indicators into a dataframe"""
    df = pd.DataFrame()

    if indicators == "all":
        indicators = obj.indicators.values()

    if isinstance(indicators, str):
        indicators = [indicators]

    if isinstance(indicators, list):

        for _ in indicators:
            if _ not in obj.indicators:
                raise ValueError(f"{_} has not been loaded or is an invalid indicator.")

        indicators = [
            obj.indicators[_] for _ in indicators if _ in list(obj.indicators)
        ]

    for _ in indicators:
        df = pd.concat([df, _], ignore_index=True)

    return df


class SDR(ImportData):
    """An object to extract SDR data

    An object to help extract and store the latest Special Drawing Rights (SDR)
    data from the IMF website: https://www.imf.org/external/np/fin/tad/extsdr1.aspx

    In order to use, create an instance of this class.
    Then, call the load_indicator method to load an indicator. Optionally specify
    the indicator to extract (holdings or allocations).

    If the data for an indicator has never been downloaded, it will be downloaded.
    If it has been downloaded, it will be loaded from disk.
    If `update_data` is set to True when creating the object, the data will be updated each time
    `load_indicators is called`. You can force an update by calling `update` if you
    want to refresh the data stored on disk.
    You can get a dataframe of the data by calling `get_data`.

    The data consists of `member` (including countries, prescribed holders e.g. central banks,
    and other holders - "General Resource Account"), `indicator` (holdings or allocations), `date`
    (the date of the SDR announcement), and `value`.
    """

    def load_indicator(self, indicator: Optional[str | list] = "all") -> ImportData:
        """Load SDR data. Optionally specify the indicator to load (holdings or allocations).

        Args:
            indicator: indicator to load, either 'holdings' or 'allocations'. The default
                is 'all' which loads both.

        Returns:
            The same object to allow chaining
        """

        # make sure indicators are valid
        indicator_list = _check_sdr_indicators(indicator)

        if not os.path.exists(f"{PATHS.imported_data}/{self.file_name}") or (
            self.update_data and self.data is None
        ):
            self.update(reload_data=False)

        if self.data is None:
            self.data = pd.read_csv(f"{PATHS.imported_data}/{self.file_name}")

        for i in indicator_list:
            self.indicators[i] = self.data[self.data["indicator"] == i].reset_index(
                drop=True
            )

        return self

    def update(self, reload_data: bool = True) -> ImportData:
        """Update the data saved on disk

        When called it extracts the SDR data from the IMF website and saves it to disk.
        Optionally specify whether to reload the data to the object

        Args:
            reload_data: If True, the data will be reloaded to the object

        Returns:
            The same object to allow chaining
        """

        base_url = "https://www.imf.org/external/np/fin/tad/"

        links_url = "https://www.imf.org/external/np/fin/tad/extsdr1.aspx"

        # check latest year
        years = _parse_sdr_links(url=links_url, concat_url=base_url)
        latest_year_link = list(years.values())[0]

        # check latest date
        dates = _parse_sdr_links(latest_year_link, base_url)
        latest_date_link = list(dates.values())[0]

        # Convert latest date to formatted string
        latest_date = datetime.strptime(list(dates.keys())[0], "%B %d, %Y").strftime(
            "%d %B %Y"
        )
        # find tsv file
        tsv_link = _get_tsv_url(latest_date_link)

        # read tsv file
        df = _read_sdr_tsv(tsv_link)

        (
            df.assign(date=latest_date).to_csv(
                f"{PATHS.imported_data}/{self.file_name}", index=False
            )
        )

        if reload_data:
            self.load_indicator(list(self.indicators))
        else:
            print(
                "Successfully updated SDR data to disk. "
                "Run `load_indicator` to load the data"
            )

        return self

    def get_data(
        self,
        indicators: str = "all",
        members: Optional[str | list] = "all",
    ) -> pd.DataFrame:
        """Get the data as a Pandas DataFrame

        Args:
            indicators: indicator to get, either 'holdings' or 'allocations'.
                If 'all', both indicators will be returned

            members: member to get (includes countries, prescribed holders, and others).
                If 'all', all members will be returned

        Returns:
            A pandas dataframe with the SDR data
        """
        df = _get_data(obj=self, indicators=indicators)

        if isinstance(members, str) and members != "all":
            if members not in self.members:
                raise ValueError(
                    f"member not found: {members}.\nPlease call `obj.member` to see available members."
                )

            df = df[df["member"] == members]

        elif isinstance(members, list):
            for member in members:
                if member not in df["member"].unique():
                    warnings.warn(
                        f"member not found: {member}.\nPlease call `obj.member` to see available members."
                    )
            df = df[df["member"].isin(members)]

        if len(df) == 0:
            raise ValueError(f"No members found")

        return df.reset_index(drop=True)

    @property
    def file_name(self):
        """Returns the name of the stored file"""
        return f"SDR.csv"

    @property
    def members(self):
        """Returns a list of all members"""
        return self.data["member"].unique()


def _check_weo_parameters(
    latest_y: int | None = None, latest_r: int | None = None
) -> (int, int):
    """Check parameters and return max values or provided input"""
    if latest_y is None:
        latest_y = max(*all_releases())[0]

    # if latest release isn't provided, take max value
    if latest_r is None:
        latest_r = max(*all_releases())[1]

    return latest_y, latest_r


def _update_weo(latest_y: int = None, latest_r: int = None) -> None:
    """Update data from the World Economic Outlook, using WEO package"""

    latest_y, latest_r = _check_weo_parameters(latest_y, latest_r)

    # Download the file from the IMF website and store in directory
    download(
        latest_y,
        latest_r,
        directory=PATHS.imported_data,
        filename=f"weo{latest_y}_{latest_r}.csv",
    )

    # Validate the file
    if os.path.getsize(f"{PATHS.imported_data}/weo{latest_y}_{latest_r}.csv") < 1000:
        print(
            f"Downloading release {latest_r} of "
            f"{latest_y} failed. Trying previous release"
        )
        os.remove(f"{PATHS.imported_data}/weo{latest_y}_{latest_r}.csv")

        try:
            _update_weo(latest_y, latest_r - 1)
        except weo.dates.DateError:
            _update_weo(latest_y - 1, latest_r)


class WorldEconomicOutlook(ImportData):
    """World Economic Outlook data"""

    year: Optional[int] = None
    release: Optional[int] = None

    def __load_data(
        self, latest_y: int | None = None, latest_r: int | None = None
    ) -> None:
        """loading WEO as a clean dataframe

        Args:
            latest_y: passed only optional to override the behaviour to get the latest
            release year for the WEO.
            latest_r: passed only optionally to override the behaviour to get the latest
            released value (1 or 2).
        """

        latest_y, latest_r = _check_weo_parameters(latest_y, latest_r)

        names = {
            "ISO": "iso_code",
            "WEO Subject Code": "indicator",
            "Subject Descriptor": "indicator_name",
            "Subject Notes": "indicator_description",
            "Units": "units",
            "Scale": "scale",
            "Estimates Start After": "estimates_start_after",
        }

        to_drop = [
            "WEO Country Code",
            "Country",
            "Country/Series-specific Notes",
        ]

        # If data doesn't exist or update is required, update the data
        if (
            not os.path.exists(f"{PATHS.imported_data}/weo{latest_y}_{latest_r}.csv")
            or self.update_data
        ):
            _update_weo(latest_y, latest_r)

        # Load the data from disk. If it doesn't exist, try the previous one
        try:
            df = WEO(PATHS.imported_data + rf"/weo{latest_y}_{latest_r}.csv").df
            self.version = {"year": latest_y, "release": latest_r}
        except FileNotFoundError:
            try:
                df = WEO(PATHS.imported_data + rf"/weo{latest_y}_{latest_r-1}.csv").df
                self.version = {"year": latest_y, "release": latest_r - 1}
            except FileNotFoundError:
                df = WEO(PATHS.imported_data + rf"/weo{latest_y-1}_{latest_r}.csv").df
                self.version = {"year": latest_y - 1, "release": latest_r}

        # Load data into data object
        self.data = (
            df.drop(to_drop, axis=1)
            .rename(columns=names)
            .melt(id_vars=names.values(), var_name="year", value_name="value")
            .assign(
                year=lambda d: pd.to_datetime(d.year, format="%Y"),
                value=lambda d: clean_numeric_series(d.value),
            )
            .dropna(subset=["value"])
            .reset_index(drop=True)
        )

    def _check_indicators(self, indicators: str | list | None = None) -> None | dict:

        if self.data is None:
            self.__load_data()

        # Create dictionary of available indicators
        indicators_ = (
            self.data.drop_duplicates(subset=["indicator", "indicator_name", "units"])
            .assign(name_units=lambda d: d.indicator_name + " (" + d.units + ")")
            .set_index("indicator")["name_units"]
            .to_dict()
        )

        if indicators is None:
            return indicators_

        if isinstance(indicators, str):
            indicators = [indicators]

        for _ in indicators:
            if _ not in indicators_:
                self.available_indicators()
                raise ValueError(f"Indicator not found: {_}")

    def load_indicator(
        self, indicator_code: str, indicator_name: Optional[str] = None
    ) -> ImportData:
        """Loads a specific indicator from the World Economic Outlook data"""

        if self.data is None:
            self.__load_data()

        # Check if indicator exists
        self._check_indicators(indicators=indicator_code)

        self.indicators[indicator_code] = (
            self.data.loc[lambda d: d.indicator == indicator_code]
            .assign(
                indicator_name=indicator_name
                if indicator_name is not None
                else lambda d: d.indicator_name,
                estimate=lambda d: d.apply(
                    lambda r: True if r.year.year >= r.estimates_start_after else False,
                    axis=1,
                ),
            )
            .drop(columns=["estimates_start_after"])
            .sort_values(["iso_code", "year"])
            .reset_index(drop=True)
        )
        return self

    def update(self, latest_y: int | None = None, latest_r: int | None = None) -> None:
        """Update the stored WEO data, using WEO package.

        Args:
            latest_y: passed only optionally to override the behaviour to get the latest
                release year for the WEO.
            latest_r: passed only optionally to override the behaviour to get the latest
                released value (1 or 2).
        """
        _update_weo(latest_y=latest_y, latest_r=latest_r)

    def available_indicators(self) -> None:
        """Print the available indicators in the dataset"""

        indicators_ = self._check_indicators(indicators=None)

        available = [f"{code}: {name} \n" for code, name in indicators_.items()]

        print(f"Available indicators:\n{''.join(available)}")

    def get_data(
        self, indicators: str | list = "all", keep_metadata: bool = False
    ) -> pd.DataFrame:

        df = _get_data(obj=self, indicators=indicators)

        if not keep_metadata:
            return df.filter(["iso_code", "name", "indicator", "year", "value"], axis=1)

        return df


def __get_rate(rows: list, text: str):
    """Returns currency value from SDR exchange rate table"""

    for i in range(len(rows)):
        if text in rows[i].text:
            return float(rows[i + 1].text.strip().split("\r\n")[0])
    else:
        raise ValueError("Could not find exchange rate")


def latest_sdr_exchange(currency: Optional | str = "USD") -> dict[str:str, str:float]:
    """Extracts the latest SDR exchange rate and date

    Args:
        currency: exchange rate currency, either "USD" or "SDR". Defaults to "USD". If "USD" is selected
                  the value returned is the value of SDRs equivalent to 1 USD

    Returns:
        dictionary with date and value
    """

    page = "https://www.imf.org/external/np/fin/data/rms_sdrv.aspx"
    exchange_info = {}

    try:
        content = requests.get(page).content
    except ConnectionError:
        raise ConnectionError("Could not extract exchange rates")

    soup = BeautifulSoup(content, "html.parser")
    table = soup.find_all("table")[5]

    date = table.find_all("th")[0].text.strip()
    date = datetime.strptime(date, "%A, %B %d, %Y").strftime("%d %b %Y")
    exchange_info.update({"date": date})

    rows = table.find_all("td")

    if currency == "USD":
        exchange_info.update({"value": __get_rate(rows, "U.S.$1.00 = SDR")})
    elif currency == "SDR":
        exchange_info.update({"value": __get_rate(rows, "SDR1 = US$")})
    else:
        raise ValueError('Invalid currency. Please select from ["SDR", "USD]')

    return exchange_info



from __future__ import annotations

import os
import warnings
from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd
import wbgapi as wb

from bblocks.config import PATHS
from bblocks.import_tools.common import ImportData


def _get_wb_data(
        series: str,
        series_name: str | None = None,
        start_year: int | None = None,
        end_year: int | None = None,
        most_recent_only: bool = False,
) -> pd.DataFrame:
    """Get data for an indicator, using wbgapi"""

    if (start_year is not None) is not (end_year is not None):
        raise ValueError("start_year and end_year must both be provided")

    time_period = (
        range(start_year, end_year + 1)
        if all([isinstance(start_year, int), isinstance(end_year, int)])
        else "all"
    )

    # get data
    return (
        wb.data.DataFrame(
            series=series,
            time=time_period,
            mrnev=1 if most_recent_only else None,
            numericTimeKeys=True,
            labels=False,
            columns="series",
            timeColumns=True,
        )
            .reset_index()
            .rename(
            columns={
                "economy": "iso_code",
                "index": "iso_code",
                series: "value",
                "time": "date",
                f"{series}:T": "date",
            }
        )
            .assign(
            indicator=series_name if series_name is not None else series,
            indicator_code=series,
            date=lambda d: pd.to_datetime(d.date, format="%Y"),
        )
            .sort_values(by=["iso_code", "date"])
            .reset_index(drop=True)
            .filter(["date", "iso_code", "indicator", "indicator_code", "value"], axis=1)
    )


class WorldBankData(ImportData):
    """An object to help download data from the World Bank.
    In order to use, create an instance of this class.
    Then, call the load_indicator method to load an indicator. This can be done multiple times.
    If the data for an indicator has never been downloaded, it will be downloaded.
    If it has been downloaded, it will be loaded from disk.
    If `update_data` is set to True when creating the object, the data will be updated
    from the World Bank for each indicator.
    You can force an update by calling `update` if you want to refresh the data stored on disk.
    You can get a dataframe of the data by calling `get_data`."""

    def load_indicator(
            self,
            indicator_code: str,
            indicator_name=None,
            start_year: Optional | int = None,
            end_year: Optional | int = None,
            most_recent_only: bool = False,
    ) -> WorldBankData:
        """Get an indicator from the World Bank API

        Args:
            indicator_code: the code from the World Bank data portal (e.g. "SP.POP.TOTL")
            indicator_name: A human-readable name for the indicator (e.g. "Total Population")
            start_year: The first year to include in the data
            end_year: The last year to include in the data
            most_recent_only: If True, only get the most recent non-empty value for each country

        Returns:
            The same object to allow chaining

        """
        years_str = (
            f"{start_year}-{end_year}"
            if all([isinstance(start_year, int), isinstance(end_year, int)])
            else "all"
        )
        file_name = (
            f"{indicator_code}_{years_str}_"
            f"{'most_recent' if most_recent_only else ''}.csv"
        )

        __params = {
            "series": indicator_code,
            "series_name": indicator_name,
            "start_year": start_year,
            "end_year": end_year,
            "most_recent_only": most_recent_only,
        }

        # get the indicator data if it's not saved on disk.
        if not os.path.exists(f"{PATHS.imported_data}/{file_name}") or self.update_data:
            _get_wb_data(**__params).to_csv(
                f"{PATHS.imported_data}/{file_name}", index=False
            )

        _ = pd.read_csv(f"{PATHS.imported_data}/{file_name}", parse_dates=["date"])

        __params["file_name"] = file_name

        self.indicators[indicator_code] = _, __params

        return self

    def update(self) -> ImportData:
        """Update the data saved on disk for the different indicators

        When called, it will go through each indicator and update the data saved
        based on the parameters passed to load_indicator.

        Returns:
            The same object to allow chaining

        """
        if len(self.indicators) == 0:
            raise RuntimeError("No indicators loaded")

        for indicator_code, (_, args) in self.indicators.items():
            file_name = args.pop("file_name")
            _get_wb_data(**args).to_csv(
                f"{PATHS.imported_data}/{file_name}", index=False
            )

        return self

    def get_data(self, indicators: Optional[str | list] = "all") -> pd.DataFrame:
        """
        Get the data as a Pandas DataFrame
        Args:
            indicators: By default, all indicators are returned in a single DataFrame.
            If a list of indicators is passed, only those indicators will be returned.
            A single indicator can be passed as a string as well.

        Returns:
            A Pandas DataFrame with the data for the indicators requested.

        """

        df = pd.DataFrame()

        if indicators != "all" and isinstance(indicators, str):
            indicators = [indicators]

        if isinstance(indicators, list):
            indicators = [
                self.indicators[_] for _ in indicators if _ in list(self.indicators)
            ]

        elif indicators == "all":
            indicators = self.indicators.values()

        for indicator in indicators:
            df = pd.concat([df, indicator[0]], ignore_index=True)

        self.data = df
        return self.data


def _clean_pink_sheet(df: pd.DataFrame, sheet: str) -> pd.DataFrame:
    """clean, format, standardize pink sheet data"""

    if sheet == "Monthly Prices":
        df.columns = df.iloc[3]
        df = (
            df.rename(columns={np.nan: "period"})
                .iloc[6:]
                .assign(period=lambda d: pd.to_datetime(d.period, format="%YM%m"))
                .replace("..", np.nan)
                .reset_index(drop=True)
        )

    elif sheet == "Monthly Indices":
        df = df.iloc[9:].replace("..", np.nan).reset_index(drop=True)
        df.columns = [
            "period",
            "Energy",
            "Non-energy",
            "Agriculture",
            "Beverages",
            "Food",
            "Oils & Meals",
            "Grains",
            "Other Food",
            "Raw Materials",
            "Timber",
            "Other Raw Mat.",
            "Fertilizers",
            "Metals & Minerals",
            "Base Metals (ex. iron ore)",
            "Precious Metals",
        ]
        df["period"] = pd.to_datetime(df.period, format="%YM%m")
    df = pd.melt(df, id_vars="period", var_name="indicator", value_name="value")

    return df


@dataclass
class PinkSheet(ImportData):
    """An object to help download data from World Bank Pink sheets.

    In order to use, create an instance of this class, specifying the sheet name - 'Monthly Prices', 'Monthly Indices'.
    Then, call the load_indicator method to load an indicator, optionally specifying in indicator or
    list of indicators. This can be done multiple times. If the data has never been downloaded,
    it will be downloaded. If it has been downloaded, it will be loaded from disk.
    If `update_data` is set to True when creating the object, the full dataset will be downloaded to disk
    when `load_indicator` is called for the first time.
    You can force an update by calling `update` if you want to refresh the data stored on disk.
    You can get a dataframe of the data by calling `get_data`

    """

    sheet: str = None

    def __post_init__(self):
        if self.sheet not in ["Monthly Indices", "Monthly Prices"]:
            raise ValueError(
                "Invalid sheet name. "
                "Please specify 'Monthly Indices' or 'Monthly Prices'"
            )

    def load_indicator(self, indicator: Optional[str | list] = "all") -> ImportData:
        """Load data for an indicator or list of indicators.

        Args:
            indicator: indicator to load, either. The default
                is 'all' which loads both.

        Returns:
            The same object to allow chaining
        """

        if not os.path.exists(f"{PATHS.imported_data}/{self.file_name}") or (
                self.update_data and self.data is None
        ):
            self.update(reload_data=False)

        if self.data is None:
            self.data = pd.read_csv(f"{PATHS.imported_data}/{self.file_name}")

        if indicator == "all":
            indicator = list(self.data.indicator.unique())
        elif isinstance(indicator, str):
            indicator = [indicator]

        for i in indicator:
            if i not in self.data.indicator.unique():
                raise Warning(f"Indicator not found: {i}")
            else:
                self.indicators[i] = (self.data[self.data.indicator == i]
                                      .reset_index(drop=True))

        return self

    def update(self, reload_data=True) -> ImportData:
        """Update the data saved on disk

        When called it downloads Pink sheet Data from the World Bank and saves it to disk.
        Optionally specify whether to reload the data to the object

        Args:
            reload_data: If True, the data will be reloaded to the object

        Returns:
            The same object to allow chaining
        """

        url = (
            "https://thedocs.worldbank.org/en/doc/5d903e848db1d1b83e0ec8f744e55570-0350012021/related/CMO"
            "-Historical-Data-Monthly.xlsx"
        )

        (
            pd.read_excel(url, sheet_name=self.sheet)
                .pipe(_clean_pink_sheet, sheet=self.sheet)
                .to_csv(f"{PATHS.imported_data}/{self.file_name}", index=False)
        )

        if reload_data:
            self.load_indicator(list(self.indicators))

        return self

    def get_data(
            self,
            indicators: Optional[str | list] = None,
            start_date: Optional[str] = None,
            end_date: Optional[str] = None,
    ) -> pd.DataFrame:
        """Get the data as a Pandas DataFrame

        Args:
            indicators: indicator to get. If 'all', both indicators will be returned
            start_date: start date of the data (e.g. '2019-01-01')
            end_date: end date of the data (e.g. '2019-12-31')

        Returns:
            Pandas DataFrame of the data
        """

        if (start_date is not None) & (end_date is not None):
            if start_date > end_date:
                raise ValueError("start date cannot be earlier than end date")

        df = pd.DataFrame()

        if indicators is not None:
            if isinstance(indicators, str):
                indicators = [indicators]
            for indicator in indicators:
                if indicator not in self.indicators:
                    warnings.warn(f"Indicator not found: {indicator}")
                else:
                    df = pd.concat([df, self.indicators[indicator]])

        else:
            df = self.data

        if start_date is not None:
            df = df[df["period"] >= start_date]
        if end_date is not None:
            df = df[df["period"] <= end_date]

        if len(df) == 0:
            raise ValueError("No data available for current parameters")

        return df.reset_index(drop=True)

    @property
    def file_name(self) -> str:
        """Returns the name of the stored file"""
        return f"World Bank Pink Sheet - {self.sheet}.csv"



import logging
import os

import camelot
import country_converter as coco
import pandas as pd
import requests
from bblocks.config import PATHS
from numpy import nan

URL: str = "https://www.imf.org/external/Pubs/ft/dsa/dsalist.pdf"


def __download_dsa_pdf(url: str, local_path: str) -> None:
    """Downloads dsa pdf to the file"""

    response = requests.get(url)

    if response.status_code != 200:
        raise ConnectionError("Could not download PDF")

    with open(local_path, "wb") as f:
        f.write(response.content)


def __pdf_to_df(local_path: str) -> pd.DataFrame:
    """Reads a pdf and returns a dataframe"""

    try:
        tables = camelot.read_pdf(local_path, flavor="stream")
        if len(tables) != 1:
            raise ValueError("Invalid PDF format. Check PDF")

        return tables[0].df

    except ValueError:
        raise ValueError("Could not read PDF to a dataframe")


def __clean_dsa(df: pd.DataFrame) -> pd.DataFrame:
    """Clean dsa dataframe"""

    logging.getLogger("country_converter").setLevel(
        logging.ERROR
    )  # silence country_converter

    columns = {0: "country", 1: "latest_publication", 2: "risk_of_debt_distress"}

    return (
        df.filter(columns.keys())
        .rename(columns=columns)
        .assign(
            country=lambda d: coco.convert(d.country, to="name_short", not_found=nan)
        )
        .dropna(subset=["country"])
        .replace({"…": nan, "": nan})
        .assign(latest_publication=lambda d: pd.to_datetime(d.latest_publication))
        .reset_index(drop=True)
    )


def get_dsa(
    update=False, local_path: str = f"{PATHS.imported_data}/dsa_list.pdf"
) -> pd.DataFrame:
    """Extract DSA data from the

    Extract the most recent Debt Sustainability Assessment (DSA) data
    for PRGT-Eligible Countries from the IMF website.
    URL = https://www.imf.org/external/Pubs/ft/dsa/DSAlist.pdf

    Args:
        local_path: where the downloaded PDF will be stored
        update (bool): if True, updates the data from the IMF website. Otherwise
            it loads the data from the local file. If a local file does not exist,
            the data will be extracted from the website.

    Returns:
        pandas dataframe with country, latest publication date, and risk of debt distress
    """

    if not os.path.exists(local_path) or update:
        __download_dsa_pdf(URL, local_path)

    return __pdf_to_df(local_path).pipe(__clean_dsa)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import pandas as pd


@dataclass
class ImportData(ABC):
    indicators: dict = field(default_factory=dict)
    data: pd.DataFrame = None
    update_data: bool = False

    @abstractmethod
    def load_indicator(self, **kwargs) -> ImportData:
        """Load the data saved on disk"""
        pass

    @abstractmethod
    def update(self, **kwargs) -> ImportData:
        """Update object data"""
        pass

    @abstractmethod
    def get_data(self, **kwargs) -> pd.DataFrame:
        """Get the data as a Pandas DataFrame"""
        pass


def append_new_data(
    new_data: pd.DataFrame, existing_data_path: str, parse_dates: list[str] | None
) -> pd.DataFrame:
    """Append new data to an existing dataframe"""
    # Read file
    try:
        saved = pd.read_csv(existing_data_path, parse_dates=parse_dates)
    except FileNotFoundError:
        saved = pd.DataFrame()

    # Append new data
    data = pd.concat([saved, new_data], ignore_index=True)

    return data.drop_duplicates(keep="last").reset_index(drop=True)



import pandas as pd

from bblocks.cleaning_tools.filter import filter_latest_by
from bblocks.import_tools.imf import WorldEconomicOutlook
from bblocks.import_tools.world_bank import WorldBankData


def __get_wb_ind(
    ind_code: str, ind_name: str, update: bool, mrnev: bool
) -> pd.DataFrame:
    """Get a simplified DataFrame for a World Bank indicator"""
    return (
        WorldBankData(update_data=update)
        .load_indicator(
            ind_code,
            most_recent_only=mrnev,
        )
        .get_data()
        .assign(year=lambda d: d.date.dt.year)
        .filter(["year", "iso_code", "value"], axis=1)
        .rename(columns={"value": ind_name})
    )


def get_population_df(*, most_recent_only, update_population_data) -> pd.DataFrame:
    """Get a population DataFrame"""

    return __get_wb_ind(
        ind_code="SP.POP.TOTL",
        ind_name="population",
        update=update_population_data,
        mrnev=most_recent_only,
    )


def get_poverty_ratio_df(*, most_recent_only, update_poverty_data) -> pd.DataFrame:
    """Get a population DataFrame"""

    return __get_wb_ind(
        ind_code="SI.POV.DDAY",
        ind_name="poverty_headcount_ratio",
        update=update_poverty_data,
        mrnev=most_recent_only,
    )


def get_population_density_df(
    *, most_recent_only, update_population_data
) -> pd.DataFrame:
    """Get a population DataFrame"""

    return __get_wb_ind(
        ind_code="EN.POP.DNST",
        ind_name="population_density",
        update=update_population_data,
        mrnev=most_recent_only,
    )


def _get_weo_indicator(
    indicator: str,
    *,
    most_recent_only: bool,
    update_data: bool,
    include_estimates: bool = True,
) -> pd.DataFrame:

    # Create a World Economic Outlook object
    weo = WorldEconomicOutlook(update_data=update_data)

    # Get the data
    data = (
        weo.load_indicator(indicator)
        .get_data(keep_metadata=True)
        .assign(year=lambda d: pd.to_datetime(d.year, format="%Y"))
    )

    # Filter the data to keep only non-estimates if needed
    if not include_estimates:
        data = data.loc[lambda d: ~d.estimate]

    # limit most recent to current year
    data = data.loc[lambda d: d.year.dt.year <= pd.Timestamp.now().year]

    # Filter the data to keep only the most recent data if needed
    if most_recent_only:
        data = filter_latest_by(
            data, date_column="year", value_columns="value", group_by=["iso_code"]
        )

    return data.assign(year=lambda d: d.year.dt.year).filter(
        ["year", "iso_code", "value"], axis=1
    )


def get_gdp_df(
    *,
    usd: bool = True,
    most_recent_only: bool,
    update_gdp_data: bool,
    include_estimates: bool = True,
) -> pd.DataFrame:
    """Get a population DataFrame"""

    # Select the right indicator depending on the USD flag
    indicator = "NGDPD" if usd else "NGDP"

    return _get_weo_indicator(
        indicator,
        most_recent_only=most_recent_only,
        update_data=update_gdp_data,
        include_estimates=include_estimates,
    ).assign(
        value=lambda d: d.value * 1e9,
    )


def get_gov_expenditure_df(
    *,
    usd: bool = True,
    most_recent_only: bool,
    update_data: bool,
    include_estimates: bool = True,
) -> pd.DataFrame:
    """Get a population DataFrame"""

    gov_exp = _get_weo_indicator(
        indicator="GGX_NGDP",
        most_recent_only=most_recent_only,
        update_data=update_data,
        include_estimates=include_estimates,
    )

    gdp = get_gdp_df(
        usd=usd,
        most_recent_only=most_recent_only,
        update_gdp_data=update_data,
        include_estimates=include_estimates,
    )

    return (
        gov_exp.merge(
            gdp, on=["year", "iso_code"], how="left", suffixes=("_gov", "_gdp")
        )
        .assign(value=lambda d: d.value_gov / 100 * d.value_gdp)
        .filter(["year", "iso_code", "value"], axis=1)
    )



import re
import warnings
from typing import Type, Optional

import country_converter as coco
import pandas as pd
from numpy import nan


def clean_number(number: str, to: Type = float) -> float | int:
    """Clean a string and return as float or integer.
    When selecting to=int, the default python round behaviour is used.

    Args:
        number: the string to be cleaned
        to: the type to convert to (int or float)

    """

    if not isinstance(number, str):
        number = str(number)

    number = re.sub(r"[^\d.]", "", number)

    if number == "":
        return nan

    if to == float:
        return float(number)

    if to == int:
        return int(round(float(number)))


def clean_numeric_series(
    data: pd.Series | pd.DataFrame,
    series_columns: Optional[str | list] = None,
    to: Type = float,
) -> pd.DataFrame | pd.Series:
    """Clean a numeric column in a Pandas DataFrame or a Pandas Series which is
    meant to be numeric. When selecting to=int, the default python round behaviour
    is used.

    Args:
        data: it accepts a series or a dataframe. If a dataframe is passed, the column(s)
            to clean must be specified
        series_columns: optionally declared (only when data is a dataframe). To apply to
            one or more columns.
        to: the type to convert to (int or float)

    """

    if isinstance(data, pd.DataFrame) and (series_columns is None):
        raise ValueError("series_column must be specified when data is a DataFrame")

    if isinstance(data, pd.DataFrame):
        if isinstance(series_columns, str):
            series_columns = [series_columns]

        data[series_columns] = data[series_columns].apply(
            lambda s: s.apply(clean_number, to=to), axis=1
        )
        return data

    if isinstance(data, pd.Series):
        return data.apply(clean_number, to=to)


def to_date_column(series: pd.Series, date_format: Optional[str] = None) -> pd.Series:
    """Converts a Pandas series into a date series.
    The series must contain integers or strings that can be converted into
    datetime objects"""

    if pd.api.types.is_datetime64_any_dtype(series):
        return series

    if pd.api.types.is_numeric_dtype(series):
        try:
            return pd.to_datetime(series, format="%Y")

        except ValueError:
            raise ValueError(
                f"could not parse date format in. "
                f"To fix, convert column to datetime"
            )
    if date_format is None:
        return pd.to_datetime(series, infer_datetime_format=True)
    else:
        return pd.to_datetime(series, format=date_format)


def convert_id(
    series: pd.Series,
    from_type: str = "regex",
    to_type: str = "ISO3",
    not_found: str | None = None,
    *,
    additional_mapping: dict = None,
) -> pd.Series:
    """Takes a Pandas' series with country IDs and converts them into the desired type.

    Args:
        series: the Pandas series to convert
        from_type: the classification type according to which the series is encoded.
            Available types come from the country_converter package
            (https://github.com/konstantinstadler/country_converter#classification-schemes)
            For example: ISO3, ISO2, name_short, DACcode, etc.
        to_type: the target classification type. Same options as from_type
        not_found: what to do if the value is not found. Can pass a string or None.
            If None, the original value is passed through.
        additional_mapping: Optionally, a dictionary with additional mappings can be used.
            The keys are the values to be converted and the values are the converted values.
            The keys follow the same datatype as the original values. The values must follow
            the same datatype as the target type.
    """

    # if from and to are the same, return without changing anything
    if from_type == to_type:
        return series

    # Create convert object
    cc = coco.CountryConverter()

    # Get the unique values for mapping. This is done in order to significantly improve
    # the performance of country_converter with very long datasets.
    s_unique = series.unique()

    # Create a correspondence dictionary
    mapping = pd.Series(
        cc.convert(names=s_unique, src=from_type, to=to_type, not_found=nan),
        index=s_unique,
    ).to_dict()

    # If additional_mapping is passed, add to the mapping
    if additional_mapping is not None:
        mapping = mapping | additional_mapping

    return series.map(mapping).fillna(series if not_found is None else not_found)


def date_to_str(series: pd.Series, date_format: str = "%d %B %Y") -> pd.Series:
    """Converts a Pandas' series into a string series.

    Args:
        series: the Pandas series to convert to a formatted date string
        date_format: the format to use for the date string. The default is "%d %B %Y"
    """

    if not pd.api.types.is_datetime64_any_dtype(series):
        try:
            series = pd.to_datetime(series, infer_datetime_format=True)
        except ValueError:
            raise ValueError(
                f"could not parse date format in. "
                f"To fix, convert column to datetime"
            )

    return series.dt.strftime(date_format)


def format_number(
    series: pd.Series,
    as_units: bool = False,
    as_percentage: bool = False,
    as_millions: bool = False,
    as_billions: bool = False,
    decimals: int = 2,
    add_sign: bool = False,
    other_format: str = "{:,.2f}",
) -> pd.Series:
    """Formats a Pandas' numeric series into a formatted string series.

    Args:
        series: the series to convert to a formatted string
        as_units: formatted with commas to separate thousands and the specified decimals
        as_percentage: formatted as a percentage with the specified decimals. This assumes
            that the series contains numbers where 1 would equal 100%.
        as_millions: divided by 1 million, formatted with commas and the specified decimals
        as_billions: divided by 1 billion, formatted with commas and the specified decimals
        decimals: the number of decimals to use
        other_format: Other formats to use. This option can only be used if all others
            are false. Examples are available at:
            https://mkaz.blog/code/python-string-format-cookbook/
    """

    if not pd.api.types.is_numeric_dtype(series):
        raise ValueError(f"The series must be of numeric type")

    if (true_params := sum([as_units, as_percentage, as_millions, as_billions])) > 1:
        raise KeyError(
            f"Only one of as_units, as_percentage, as_millions, as_billions can be True"
        )
    if true_params > 0 and other_format != "{:,.2f}":
        raise KeyError(f"other_format can only be used if all other options are False")

    if true_params == 0:
        warnings.showwarning(
            "Using 'other_format'. The decimals setting is ignored.",
            UserWarning,
            "clean.py",
            168,
        )

    sign = "-" if add_sign else ""

    formats = {
        "as_units": f"{{:{sign},.{decimals}f}}",
        "as_percentage": f"{{:{sign},.{decimals}%}}",
        "as_millions": f"{{:{sign},.{decimals}f}}",
        "as_billions": f"{{:{sign},.{decimals}f}}",
    }

    if as_units:
        return series.map(formats["as_units"].format)

    if as_percentage:
        return series.map(formats["as_percentage"].format)

    if as_millions:
        series = series / 1e6
        return series.map(formats["as_millions"].format)

    if as_billions:
        series = series / 1e9
        return series.map(formats["as_billions"].format)

    return series.map(f"{other_format}".format)



from typing import Optional

import pandas as pd

from bblocks.analysis_tools.get import __validate_cols
from bblocks.cleaning_tools import clean
from bblocks.cleaning_tools.clean import convert_id


def filter_latest_by(
    data: pd.DataFrame,
    date_column: str,
    value_columns: str | list = None,
    group_by: Optional[str | list] = None,
) -> pd.DataFrame:
    """
    Calculate the latest value of (a) column(s) over a period of time.

    Args:
        data: a DataFrame with a date column (datetime or int) and one or more numeric columns
        date_column: the name of the date (datetime or int) column
        value_columns: one or more columns to calculate the average over
        group_by: Optionally, specify which columns to consider for the latest operation
    Returns:
        A DataFrame with the latest value of the specified columns
    """

    # Create a copy of the dataframe to avoid modifying the original
    data = data.copy(deep=True)

    # columns order
    col_order = data.columns

    # Validate args
    _, _, date_column, value_columns, group_by = __validate_cols(
        data,
        sdate=None,
        edate=None,
        date_col=date_column,
        value_col=value_columns,
        grouper=group_by,
    )

    data["_date_"] = clean.to_date_column(data[date_column])

    cols = [date_column] + value_columns

    return (
        data.sort_values(by=["_date_"])
        .groupby(by=group_by)[cols]
        .last()
        .reset_index()
        .filter(col_order, axis=1)
    )


def _filter_by(
    df: pd.DataFrame,
    by: str,
    by_value: str,
    id_column: str = "iso_code",
    id_type: str = "regex",
) -> pd.DataFrame:
    """
    Helper function to filter a DataFrame by membership to a specific grouping.
    The groupings come from those available through the `country_converter` package.
    More info available at:
    https://github.com/konstantinstadler/country_converter#classification-schemes

    Args:
        df: the DataFrame to filter
        by: the type of grouping to filter by (e.g. "Continent", "UNRegion", "EU")
        by_value: the value of the grouping to filter by (e.g. "Africa", "Europe", "EU")
        id_column: the name of the column to use for the id (default: "iso_code")
        id_type: the type of id to use (default: "regex")

    Returns:
        A filtered copy of the DataFrame.

    """
    if id_column not in df.columns:
        raise ValueError(f"{id_column} not in columns")

    df_ = df.copy(deep=True)

    df_["id_"] = convert_id(series=df_[id_column], from_type=id_type, to_type=by)

    return (
        df_.loc[lambda d: d.id_ == by_value]
        .drop(columns=["id_"])
        .reset_index(drop=True)
    )


def filter_by_continent(
    df: pd.DataFrame,
    continent: str,
    id_column: str = "iso_code",
    id_type: str = "regex",
) -> pd.DataFrame:
    """
    Filter a DataFrame by continent.
    Args:
        df: the DataFrame to filter
        continent: the continent to filter by (e.g. "Africa", "Europe", "EU")
        id_column: the name of the column to use for the id (default: "iso_code")
        id_type: the type of id to use (default: "regex")

    Returns:
        A filtered copy of the DataFrame.

    """
    return _filter_by(
        df=df, by="Continent", by_value=continent, id_column=id_column, id_type=id_type
    )


def filter_by_un_region(
    df: pd.DataFrame,
    region: str,
    id_column: str = "iso_code",
    id_type: str = "regex",
) -> pd.DataFrame:
    """
    Filter a DataFrame by UN region. This includes, for example, "Western Africa",
    "Eastern Africa", "Southern Asia", "Northern America", "Central America", "Eastern Asia".

    Args:
        df: the DataFrame to filter
        region: the region to filter by (e.g. "Western Africa", "Eastern Africa", etc.)
        id_column: the name of the column to use for the id (default: "iso_code")
        id_type: the type of id to use (default: "regex")

    Returns:

    """
    return _filter_by(
        df=df, by="UNRegion", by_value=region, id_column=id_column, id_type=id_type
    )


def filter_african_countries(
    df: pd.DataFrame,
    id_column: str = "iso_code",
    id_type: str = "regex",
) -> pd.DataFrame:
    """
    Filter a DataFrame to keep only African countries.
    Args:
        df: the DataFrame to filter
        id_column: the name of the column to use for the id (default: "iso_code")
        id_type: the type of id to use (default: "regex")

    Returns:
        A filtered copy of the DataFrame.

    """
    return filter_by_continent(
        df=df, continent="Africa", id_column=id_column, id_type=id_type
    )


def filter_eu_countries(
    df: pd.DataFrame,
    id_column: str = "iso_code",
    id_type: str = "regex",
) -> pd.DataFrame:
    """
    Filter a DataFrame to keep only European countries. The current list of members
    of the European Union is always used.

    Args:
        df: the DataFrame to filter
        id_column: the name of the column to use for the id (default: "iso_code")
        id_type: the type of id to use (default: "regex")

    Returns:
        A filtered copy of the DataFrame.

    """
    return _filter_by(
        df=df, by="EU", by_value="EU", id_column=id_column, id_type=id_type
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bblocks',
 'bblocks.analysis_tools',
 'bblocks.cleaning_tools',
 'bblocks.dataframe_tools',
 'bblocks.import_tools',
 'bblocks.import_tools.debt',
 'bblocks.other_tools']

package_data = \
{'': ['*'], 'bblocks.import_tools': ['stored_data/*', 'stored_data/wfp_raw/*']}

install_requires = \
['beautifulsoup4>=4.11.1,<5.0.0',
 'camelot-py>=0.10.1,<0.11.0',
 'country-converter>=0.7.7,<0.8.0',
 'numpy>=1.23,<2.0',
 'opencv-python>=4.6.0,<5.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pandas>=1.4,<2.0',
 'requests>=2.28.1,<3.0.0',
 'wbgapi==1.0.12',
 'weo>=0.7.4,<0.8.0']

setup_kwargs = {
    'name': 'bblocks',
    'version': '0.2.4',
    'description': 'A package with tools to download and analyse international development data.',
    'long_description': 'The bblocks package\n===================\n\n|pypi| |python| |Documentation| |codecov|\n\n\n**bblocks** is a python package with tools to download and analyse\ndevelopment data. These tools are meant to be the *building blocks* of\nfurther analysis.\n\n**bblocks** is currently in active development. Functionality and APIs\nare very likely to change. We welcome feedback, feature requests, and\ncollaboration. We hope that this will be a valuable resource for\norganisations working with sustainable development data.\n\n-  Documentation: https://bblocks.readthedocs.io/\n-  GitHub: https://github.com/ONECampaign/bblocks\n-  PyPI: https://pypi.org/project/bblocks/\n-  Free software: MIT\n\nInstallation\n------------\n\nbblocks can be installed from PyPI: from the command line:\n\n.. code-block:: python\n\n   pip install bblocks\n\nAlternatively, the source code is available on\n`GitHub <https://github.com/ONECampaign/bblocks>`__.\n\nBasic usage\n-----------\n\nbblocks is in active development. Check back regularly for new features.\nWe will be adding example jupyter notebooks for some of the main\nfeatures. Be sure to check the issues or projects section to learn more\nabout the features that we are planning to add.\n\nQuestions?\n----------\n\nPlease feel free to reach out via GitHub or at data at one.org\n\n.. |pypi| image:: https://img.shields.io/pypi/v/bblocks.svg\n   :target: https://pypi.org/project/bblocks/\n.. |python| image:: https://img.shields.io/pypi/pyversions/bblocks.svg\n   :target: https://pypi.org/project/bblocks/\n.. |codecov| image:: https://codecov.io/gh/ONECampaign/bblocks/branch/main/graph/badge.svg?token=YN8S1719NH\n   :target: https://codecov.io/gh/ONECampaign/bblocks\n.. |Documentation| image:: https://readthedocs.org/projects/bblocks/badge/?version=latest\n   :target: https://bblocks.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n',
    'author': 'The ONE Campaign',
    'author_email': 'data@one.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ONECampaign/bblocks',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
